import socket

import pytest

# import uvloop  # TODO: run the entire test suite with both uvloop and the vanilla loop.


@pytest.fixture(scope="session")
def unused_port():
    def f():
        with socket.socket(socket.AF_INET, socket.SOCK_STREAM) as s:
            s.bind(("127.0.0.1", 0))
            return s.getsockname()[1]

    return f



import asyncio
from typing import Optional

import click
import requests
import uvloop
from aiohttp import web

import aiomonitor
from aiomonitor.termui.commands import (
    auto_command_done,
    custom_help_option,
    monitor_cli,
)


async def simple(request: web.Request) -> web.Response:
    await asyncio.sleep(10)
    return web.Response(text="Simple answer")


async def hello(request: web.Request) -> web.StreamResponse:
    resp = web.StreamResponse()
    name = request.match_info.get("name", "Anonymous")
    answer = ("Hello, " + name).encode("utf8")
    resp.content_length = len(answer)
    resp.content_type = "text/plain"
    await resp.prepare(request)
    await asyncio.sleep(10)
    await resp.write(answer)
    await resp.write_eof()
    return resp


@monitor_cli.command(name="hello")
@click.argument("name", optional=True)
@custom_help_option
@auto_command_done
def do_hello(ctx: click.Context, name: Optional[str] = None) -> None:
    """Using the /hello GET interface

    There is one optional argument, "name".  This name argument must be
    provided with proper URL escape codes, like %20 for spaces.
    """
    name = "" if name is None else "/" + name
    r = requests.get("http://localhost:8090/hello" + name)
    click.echo(r.text + "\n")


async def main() -> None:
    loop = asyncio.get_running_loop()
    app = web.Application()
    app.router.add_get("/simple", simple)
    app.router.add_get("/hello/{name}", hello)
    app.router.add_get("/hello", hello)
    host, port = "localhost", 8090
    with aiomonitor.start_monitor(loop, locals=locals()):
        web.run_app(app, port=port, host=host)


if __name__ == "__main__":
    uvloop.install()
    try:
        asyncio.run(main())
    except KeyboardInterrupt:
        pass



import asyncio
import base64
import functools
import struct
import sys
import time
import traceback
import weakref
from asyncio.coroutines import _format_coroutine  # type: ignore
from typing import Any, Callable, Coroutine, List, Optional, TypeVar

import janus
from typing_extensions import ParamSpec

from .types import CancellationChain, TerminatedTaskInfo
from .utils import _extract_stack_from_frame

__all__ = (
    "TracedTask",
    "preserve_termination_log",
    "persistent_coro",
)

persistent_coro: "weakref.WeakSet[Coroutine]" = weakref.WeakSet()

T = TypeVar("T")
P = ParamSpec("P")


class TracedTask(asyncio.Task):
    _orig_coro: Coroutine[Any, Any, Any]
    _termination_stack: Optional[List[traceback.FrameSummary]]

    def __init__(
        self,
        *args,
        termination_info_queue: janus._SyncQueueProxy[TerminatedTaskInfo],
        cancellation_chain_queue: janus._SyncQueueProxy[CancellationChain],
        persistent: bool = False,
        **kwargs,
    ) -> None:
        if sys.version_info < (3, 11):
            kwargs.pop("context")
        super().__init__(*args, **kwargs)
        self._termination_info_queue = termination_info_queue
        self._cancellation_chain_queue = cancellation_chain_queue
        self._started_at = time.perf_counter()
        self._termination_stack = None
        self.add_done_callback(self._trace_termination)
        self._persistent = persistent

    def get_trace_id(self) -> str:
        h = hash(
            (
                id(self),
                self.get_name(),
            )
        )
        b = struct.pack("P", h)
        return base64.b32encode(b).rstrip(b"=").decode()

    def _trace_termination(self, _: "asyncio.Task[Any]") -> None:
        self_id = self.get_trace_id()
        exc_repr = (
            repr(self.exception())
            if not self.cancelled() and self.exception()
            else None
        )
        task_info = TerminatedTaskInfo(
            self_id,
            name=self.get_name(),
            coro=_format_coroutine(self._orig_coro).partition(" ")[0],
            started_at=self._started_at,
            terminated_at=time.perf_counter(),
            cancelled=self.cancelled(),
            termination_stack=self._termination_stack,
            canceller_stack=None,
            exc_repr=exc_repr,
            persistent=self._persistent,
        )
        self._termination_info_queue.put_nowait(task_info)

    def cancel(self, msg: Optional[str] = None) -> bool:
        try:
            canceller_task = asyncio.current_task()
        except RuntimeError:
            canceller_task = None
        if canceller_task is not None and isinstance(canceller_task, TracedTask):
            canceller_stack = _extract_stack_from_frame(sys._getframe())[:-1]
            cancellation_chain = CancellationChain(
                self.get_trace_id(),
                canceller_task.get_trace_id(),
                canceller_stack,
            )
            self._cancellation_chain_queue.put_nowait(cancellation_chain)
        return super().cancel(msg)


def preserve_termination_log(
    corofunc: Callable[P, Coroutine[Any, None, T]],
) -> Callable[P, Coroutine[Any, None, T]]:
    """
    Guard the given coroutine function from being stripped out due to the max history
    limit when created as TracedTask.
    """

    @functools.wraps(corofunc)
    def inner(*args: P.args, **kwargs: P.kwargs) -> Coroutine[Any, None, T]:
        coro = corofunc(*args, **kwargs)
        persistent_coro.add(coro)
        return coro

    return inner



from __future__ import annotations

import asyncio
import signal
from typing import TYPE_CHECKING, Iterable, List

import click
from click.parser import split_arg_string
from click.shell_completion import CompletionItem, _resolve_context, _resolve_incomplete
from prompt_toolkit.completion import CompleteEvent, Completer, Completion
from prompt_toolkit.document import Document

from ..context import current_monitor

if TYPE_CHECKING:
    from ..monitor import Monitor


class ClickCompleter(Completer):
    def __init__(self, root_command: click.BaseCommand) -> None:
        self._root_command = root_command

    def get_completions(
        self,
        document: Document,
        complete_event: CompleteEvent,
    ) -> Iterable[Completion]:
        args = split_arg_string(document.current_line)
        incomplete = document.get_word_under_cursor()
        if incomplete and args and args[-1] == incomplete:
            args.pop()
        click_ctx = _resolve_context(self._root_command, {}, "", args)
        cmd_or_param, incomplete = _resolve_incomplete(click_ctx, args, incomplete)
        completions: List[CompletionItem] = cmd_or_param.shell_complete(
            click_ctx, incomplete
        )
        start_position = document.find_backwards(incomplete) or 0
        for completion in completions:
            yield Completion(
                completion.value,
                start_position=start_position,
            )


def complete_task_id(
    ctx: click.Context,
    param: click.Parameter,
    incomplete: str,
) -> Iterable[str]:
    # ctx here is created in the completer and does not have ctx.obj set as monitor.
    # We take the monitor instance from the global context variable instead.
    try:
        self: Monitor = current_monitor.get()
    except LookupError:
        return []
    return [
        task_id
        for task_id in map(
            str, sorted(map(id, asyncio.all_tasks(loop=self._monitored_loop)))
        )
        if task_id.startswith(incomplete)
    ][:10]


def complete_trace_id(
    ctx: click.Context,
    param: click.Parameter,
    incomplete: str,
) -> Iterable[str]:
    try:
        self: Monitor = current_monitor.get()
    except LookupError:
        return []
    return [
        trace_id
        for trace_id in map(str, sorted(self._terminated_tasks.keys()))
        if trace_id.startswith(incomplete)
    ][:10]


def complete_signal_names(
    ctx: click.Context,
    param: click.Parameter,
    incomplete: str,
) -> Iterable[str]:
    return [sig.name for sig in signal.Signals if sig.name.startswith(incomplete)]



from __future__ import annotations

import asyncio
from contextvars import ContextVar
from typing import TYPE_CHECKING, TextIO

if TYPE_CHECKING:
    from .monitor import Monitor

current_monitor: ContextVar[Monitor] = ContextVar("current_monitor")
current_stdout: ContextVar[TextIO] = ContextVar("current_stdout")
command_done: ContextVar[asyncio.Event] = ContextVar("command_done")



from __future__ import annotations

import asyncio
import functools
import logging
import os
import shlex
import signal
import sys
import textwrap
import traceback
from contextvars import copy_context
from typing import TYPE_CHECKING, List, TextIO, Tuple

import click
from prompt_toolkit import PromptSession
from prompt_toolkit.application.current import get_app_session
from prompt_toolkit.contrib.telnet.server import TelnetConnection
from prompt_toolkit.formatted_text import FormattedText
from prompt_toolkit.shortcuts import print_formatted_text
from terminaltables import AsciiTable

from .. import console
from ..context import command_done, current_monitor, current_stdout
from ..exceptions import MissingTask
from .completion import (
    ClickCompleter,
    complete_signal_names,
    complete_task_id,
    complete_trace_id,
)

if TYPE_CHECKING:
    from ..monitor import Monitor

log = logging.getLogger(__name__)

__all__ = (
    "interact",
    "monitor_cli",
    "auto_command_done",
    "auto_async_command_done",
    "custom_help_option",
)


def _get_current_stdout() -> TextIO:
    stdout = current_stdout.get(None)
    if stdout is None:
        return sys.stdout
    else:
        return stdout


def _get_current_stderr() -> TextIO:
    stdout = current_stdout.get(None)
    if stdout is None:
        return sys.stderr
    else:
        return stdout


click.utils._default_text_stdout = _get_current_stdout
click.utils._default_text_stderr = _get_current_stderr


def print_ok(msg: str) -> None:
    print_formatted_text(
        FormattedText(
            [
                ("ansibrightgreen", "✓ "),
                ("", msg),
            ]
        )
    )


def print_fail(msg: str) -> None:
    print_formatted_text(
        FormattedText(
            [
                ("ansibrightred", "✗ "),
                ("", msg),
            ]
        )
    )


async def interact(self: Monitor, connection: TelnetConnection) -> None:
    """
    The interactive loop for each telnet client connection.
    """
    await asyncio.sleep(0.3)  # wait until telnet negotiation is done
    tasknum = len(asyncio.all_tasks(loop=self._monitored_loop))  # TODO: refactor
    s = "" if tasknum == 1 else "s"
    intro = (
        f"\nAsyncio Monitor: {tasknum} task{s} running\n"
        f"Type help for available commands\n"
    )
    print(intro, file=connection.stdout)

    # Override the Click's stdout/stderr reference cache functions
    # to let them use the correct stdout handler.
    current_monitor_token = current_monitor.set(self)
    current_stdout_token = current_stdout.set(connection.stdout)
    # NOTE: prompt_toolkit's all internal console output automatically uses
    #       an internal contextvar to keep the stdout consistent with the
    #       current telnet connection.
    prompt_session: PromptSession[str] = PromptSession(
        completer=ClickCompleter(monitor_cli),
        complete_while_typing=False,
    )
    lastcmd = "noop"
    style_prompt = "#5fd7ff bold"
    try:
        while True:
            try:
                user_input = (
                    await prompt_session.prompt_async(
                        FormattedText(
                            [
                                (style_prompt, self.prompt),
                            ]
                        )
                    )
                ).strip()
            except KeyboardInterrupt:
                print_fail("To terminate, press Ctrl+D or type 'exit'.")
            except (EOFError, asyncio.CancelledError):
                return
            except Exception:
                print_fail(traceback.format_exc())
            else:
                command_done_event = asyncio.Event()
                command_done_token = command_done.set(command_done_event)
                try:
                    if not user_input and lastcmd is not None:
                        user_input = lastcmd
                    args = shlex.split(user_input)
                    term_size = prompt_session.output.get_size()
                    ctx = copy_context()
                    ctx.run(
                        monitor_cli.main,
                        args,
                        prog_name="",
                        obj=self,
                        standalone_mode=False,  # type: ignore
                        max_content_width=term_size.columns,
                    )
                    await command_done_event.wait()
                    if args[0] == "console":
                        lastcmd = "noop"
                    else:
                        lastcmd = user_input
                except (click.BadParameter, click.UsageError) as e:
                    print_fail(str(e))
                except asyncio.CancelledError:
                    return
                except Exception:
                    print_fail(traceback.format_exc())
                finally:
                    command_done.reset(command_done_token)
    finally:
        current_stdout.reset(current_stdout_token)
        current_monitor.reset(current_monitor_token)


class AliasGroupMixin(click.Group):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self._commands = {}
        self._aliases = {}

    def command(self, *args, **kwargs):
        aliases = kwargs.pop("aliases", [])
        decorator = super().command(*args, **kwargs)

        def _decorator(f):
            cmd = decorator(click.pass_context(f))
            if aliases:
                self._commands[cmd.name] = aliases
                for alias in aliases:
                    self._aliases[alias] = cmd.name
            return cmd

        return _decorator

    def group(self, *args, **kwargs):
        aliases = kwargs.pop("aliases", [])
        # keep the same class type
        kwargs["cls"] = type(self)
        decorator = super().group(*args, **kwargs)
        if not aliases:
            return decorator

        def _decorator(f):
            cmd = decorator(f)
            if aliases:
                self._commands[cmd.name] = aliases
                for alias in aliases:
                    self._aliases[alias] = cmd.name
            return cmd

        return _decorator

    def get_command(self, ctx, cmd_name):
        if cmd_name in self._aliases:
            cmd_name = self._aliases[cmd_name]
        command = super().get_command(ctx, cmd_name)
        if command:
            return command

    def format_commands(self, ctx, formatter):
        commands = []
        for subcommand in self.list_commands(ctx):
            cmd = self.get_command(ctx, subcommand)
            # What is this, the tool lied about a command. Ignore it
            if cmd is None:
                continue
            if cmd.hidden:
                continue
            if subcommand in self._commands:
                aliases = ",".join(sorted(self._commands[subcommand]))
                subcommand = "{0} ({1})".format(subcommand, aliases)
            commands.append((subcommand, cmd))

        # allow for 3 times the default spacing
        if len(commands):
            limit = formatter.width - 6 - max(len(cmd[0]) for cmd in commands)
            rows = []
            for subcommand, cmd in commands:
                help = cmd.get_short_help_str(limit)
                rows.append((subcommand, help))
            if rows:
                with formatter.section("Commands"):
                    formatter.write_dl(rows)


@click.group(cls=AliasGroupMixin, add_help_option=False)
def monitor_cli():
    """
    To see the usage of each command, run them with "--help" option.
    """
    pass


def auto_command_done(cmdfunc):
    @functools.wraps(cmdfunc)
    def _inner(ctx: click.Context, *args, **kwargs):
        command_done_event = command_done.get()
        try:
            return cmdfunc(ctx, *args, **kwargs)
        finally:
            command_done_event.set()

    return _inner


def auto_async_command_done(cmdfunc):
    @functools.wraps(cmdfunc)
    async def _inner(ctx: click.Context, *args, **kwargs):
        command_done_event = command_done.get()
        command_done_event.clear()
        try:
            return await cmdfunc(ctx, *args, **kwargs)
        finally:
            command_done_event.set()

    return _inner


def custom_help_option(cmdfunc):
    """
    A custom help option to ensure setting `command_done_event`.
    """

    @auto_command_done
    def show_help(ctx: click.Context, param: click.Parameter, value: bool) -> None:
        if not value:
            return
        click.echo(ctx.get_help(), color=ctx.color)
        ctx.exit()

    return click.option(
        "--help",
        is_flag=True,
        expose_value=False,
        is_eager=True,
        callback=show_help,
        help="Show the help message",
    )(cmdfunc)


@monitor_cli.command(name="noop", hidden=True)
@auto_command_done
def do_noop(ctx: click.Context) -> None:
    pass


@monitor_cli.command(name="help", aliases=["?", "h"])
@custom_help_option
@auto_command_done
def do_help(ctx: click.Context) -> None:
    """Show the list of commands"""
    click.echo(monitor_cli.get_help(ctx))


@monitor_cli.command(name="signal")
@click.argument("signame", type=str, shell_complete=complete_signal_names)
@custom_help_option
@auto_command_done
def do_signal(ctx: click.Context, signame: str) -> None:
    """Send a Unix signal"""
    if hasattr(signal, signame):
        os.kill(os.getpid(), getattr(signal, signame))
        print_ok(f"Sent signal to {signame} PID {os.getpid()}")
    else:
        print_fail(f"Unknown signal {signame}")


@monitor_cli.command(name="stacktrace", aliases=["st", "stack"])
@custom_help_option
@auto_command_done
def do_stacktrace(ctx: click.Context) -> None:
    """Print a stack trace from the event loop thread"""
    self: Monitor = ctx.obj
    stdout = _get_current_stdout()
    tid = self._event_loop_thread_id
    assert tid is not None
    frame = sys._current_frames()[tid]
    traceback.print_stack(frame, file=stdout)


@monitor_cli.command(name="cancel", aliases=["ca"])
@click.argument("taskid", shell_complete=complete_task_id)
@custom_help_option
def do_cancel(ctx: click.Context, taskid: str) -> None:
    """Cancel an indicated task"""
    self: Monitor = ctx.obj

    @auto_async_command_done
    async def _do_cancel(ctx: click.Context) -> None:
        try:
            await self.cancel_monitored_task(taskid)
            print_ok(f"Cancelled task {taskid}")
        except ValueError as e:
            print_fail(repr(e))

    task = self._ui_loop.create_task(_do_cancel(ctx))
    self._termui_tasks.add(task)


@monitor_cli.command(name="exit", aliases=["q", "quit"])
@custom_help_option
@auto_command_done
def do_exit(ctx: click.Context) -> None:
    """Leave the monitor client session"""
    raise asyncio.CancelledError("exit by user")


@monitor_cli.command(name="console")
@custom_help_option
def do_console(ctx: click.Context) -> None:
    """Switch to async Python REPL"""
    self: Monitor = ctx.obj
    if not self._console_enabled:
        print_fail("Python console is disabled for this session!")
        return

    @auto_async_command_done
    async def _console(ctx: click.Context) -> None:
        log.info("Starting aioconsole at %s:%d", self._host, self._console_port)
        app_session = get_app_session()
        server = await console.start(
            self._host,
            self._console_port,
            self.console_locals,
            self._monitored_loop,
        )
        try:
            await console.proxy(
                app_session.input,
                app_session.output,
                self._host,
                self._console_port,
            )
        except asyncio.CancelledError:
            raise
        finally:
            await console.close(server, self._monitored_loop)
            log.info("Terminated aioconsole at %s:%d", self._host, self._console_port)
            print_ok("The console session is closed.")

    # Since we are already inside the UI's event loop,
    # spawn the async command function as a new task and let it
    # set `command_done_event` internally.
    task = self._ui_loop.create_task(_console(ctx))
    self._termui_tasks.add(task)


@monitor_cli.command(name="ps", aliases=["p"])
@click.option("-f", "--filter", "filter_", help="filter by coroutine or task name")
@click.option("-p", "--persistent", is_flag=True, help="show only persistent tasks")
@custom_help_option
@auto_command_done
def do_ps(
    ctx: click.Context,
    filter_: str,
    persistent: bool,
) -> None:
    """Show task table"""
    headers = (
        "Task ID",
        "State",
        "Name",
        "Coroutine",
        "Created Location",
        "Since",
    )
    self: Monitor = ctx.obj
    stdout = _get_current_stdout()
    table_data: List[Tuple[str, str, str, str, str, str]] = [headers]
    tasks = self.format_running_task_list(filter_, persistent)
    for task in tasks:
        table_data.append(
            (
                task.task_id,
                task.state,
                task.name,
                task.coro,
                task.created_location,
                task.since,
            )
        )
    table = AsciiTable(table_data)
    table.inner_row_border = False
    table.inner_column_border = False
    if filter_ or persistent:
        stdout.write(
            f"{len(tasks)} tasks running (showing {len(table_data) - 1} tasks)\n"
        )
    else:
        stdout.write(f"{len(tasks)} tasks running\n")
    stdout.write(table.table)
    stdout.write("\n")
    stdout.flush()


@monitor_cli.command(name="ps-terminated", aliases=["pt", "pst"])
@click.option("-f", "--filter", "filter_", help="filter by coroutine or task name")
@click.option("-p", "--persistent", is_flag=True, help="show only persistent tasks")
@custom_help_option
@auto_command_done
def do_ps_terminated(
    ctx: click.Context,
    filter_: str,
    persistent: bool,
) -> None:
    """List recently terminated/cancelled tasks"""
    headers = (
        "Trace ID",
        "Name",
        "Coro",
        "Since Started",
        "Since Terminated",
    )
    self: Monitor = ctx.obj
    stdout = _get_current_stdout()
    table_data: List[Tuple[str, str, str, str, str]] = [headers]
    tasks = self.format_terminated_task_list(filter_, persistent)
    for task in tasks:
        table_data.append(
            (
                task.task_id,
                task.name,
                task.coro,
                task.started_since,
                task.terminated_since,
            )
        )
    table = AsciiTable(table_data)
    table.inner_row_border = False
    table.inner_column_border = False
    if filter_ or persistent:
        stdout.write(
            f"{len(tasks)} tasks terminated " f"(showing {len(table_data) - 1} tasks)\n"
        )
    else:
        stdout.write(f"{len(tasks)} tasks terminated (old ones may be stripped)\n")
    stdout.write(table.table)
    stdout.write("\n")
    stdout.flush()


@monitor_cli.command(name="where", aliases=["w"])
@click.argument("taskid", shell_complete=complete_task_id)
@custom_help_option
@auto_command_done
def do_where(ctx: click.Context, taskid: str) -> None:
    """Show stack frames and the task creation chain of a task"""
    self: Monitor = ctx.obj
    stdout = _get_current_stdout()
    try:
        formatted_stack_list = self.format_running_task_stack(taskid)
    except MissingTask as e:
        print_fail(f"No task {e.task_id}")
        return
    for item_type, item_text in formatted_stack_list:
        if item_type == "header":
            stdout.write("\n")
            print_formatted_text(
                FormattedText(
                    [
                        ("ansiwhite", item_text),
                    ]
                )
            )
        else:
            stdout.write(textwrap.indent(item_text.strip("\n"), "  "))
            stdout.write("\n")


@monitor_cli.command(name="where-terminated", aliases=["wt"])
@click.argument("trace_id", shell_complete=complete_trace_id)
@custom_help_option
@auto_command_done
def do_where_terminated(ctx: click.Context, trace_id: str) -> None:
    """Show stack frames and the termination/cancellation chain of a task"""
    self: Monitor = ctx.obj
    stdout = _get_current_stdout()
    try:
        formatted_stack_list = self.format_terminated_task_stack(trace_id)
    except MissingTask as e:
        print_fail(f"No task {e.task_id}")
        return
    for item_type, item_text in formatted_stack_list:
        if item_type == "header":
            stdout.write("\n")
            print_formatted_text(
                FormattedText(
                    [
                        ("ansiwhite", item_text),
                    ]
                )
            )
        else:
            stdout.write(textwrap.indent(item_text.strip("\n"), "  "))
            stdout.write("\n")



import asyncio

import aiomonitor


async def main() -> None:
    loop = asyncio.get_running_loop()
    with aiomonitor.start_monitor(loop=loop) as monitor:
        print("Now you can connect with:")
        print(f"  python -m aiomonitor.cli -H {monitor.host} -p {monitor.port}")
        print("or")
        print(f"  telnet {monitor.host} {monitor.port}  # Linux only")
        loop.run_forever()


if __name__ == "__main__":
    try:
        asyncio.run(main())
    except KeyboardInterrupt:
        pass



import asyncio

import aiomonitor


async def main():
    loop = asyncio.get_running_loop()
    data = {
        "a": 123,
        "b": 456,
    }
    with aiomonitor.start_monitor(loop=loop) as monitor:
        monitor.console_locals["data"] = data
        print("Now you can connect with:")
        print(f"  python -m aiomonitor.cli -H {monitor.host} -p {monitor.port}")
        print("or")
        print(f"  telnet {monitor.host} {monitor.port}  # Linux only")
        print("\nTry 'console' command and inspect 'data' variable.")
        # run forever until interrupted
        while True:
            await asyncio.sleep(60)


if __name__ == "__main__":
    try:
        asyncio.run(main())
    except KeyboardInterrupt:
        pass



from __future__ import annotations

import asyncio
import inspect
import linecache
import sys
import traceback
from asyncio.coroutines import _format_coroutine  # type: ignore
from datetime import timedelta
from pathlib import Path
from types import FrameType
from typing import Any, List, Set

from .types import TerminatedTaskInfo


def _format_task(task: "asyncio.Task[Any]") -> str:
    """
    A simpler version of task's repr()
    """
    coro = _format_coroutine(task.get_coro()).partition(" ")[0]
    return f"<Task name={task.get_name()} coro={coro}>"


def _format_terminated_task(tinfo: TerminatedTaskInfo) -> str:
    """
    A simpler version of task's repr()
    """
    status = []
    if tinfo.cancelled:
        status.append("cancelled")
    if tinfo.exc_repr and not tinfo.cancelled:
        status.append(f"exc={tinfo.exc_repr}")
    return f"<TerminatedTask name={tinfo.name} coro={tinfo.coro} {' '.join(status)}>"


def _format_filename(filename: str) -> str:
    """
    Simplifies the site-pkg directory path of the given source filename.
    """
    stdlib = (
        f"{sys.prefix}/lib/python{sys.version_info.major}.{sys.version_info.minor}/"
    )
    site_pkg = f"{sys.prefix}/lib/python{sys.version_info.major}.{sys.version_info.minor}/site-packages/"
    home = f"{Path.home()}/"
    cwd = f"{Path.cwd()}/"
    if filename.startswith(site_pkg):
        return "<sitepkg>/" + filename[len(site_pkg) :]
    if filename.startswith(stdlib):
        return "<stdlib>/" + filename[len(stdlib) :]
    if filename.startswith(cwd):
        return "<cwd>/" + filename[len(cwd) :]
    if filename.startswith(home):
        return "<home>/" + filename[len(home) :]
    return filename


def _format_timedelta(td: timedelta) -> str:
    seconds = int(td.total_seconds())
    periods = [
        ("y", 60 * 60 * 24 * 365),
        ("m", 60 * 60 * 24 * 30),
        ("d", 60 * 60 * 24),
        ("h", 60 * 60),
        (":", 60),
        ("", 1),
    ]
    parts = []
    for period_name, period_seconds in periods:
        period_value, seconds = divmod(seconds, period_seconds)
        if period_name in (":", ""):
            parts.append(f"{period_value:02d}{period_name}")
        else:
            if period_value == 0:
                continue
            parts.append(f"{period_value}{period_name}")
    parts.append(f"{td.microseconds / 1e6:.03f}"[1:])
    return "".join(parts)


def _filter_stack(
    stack: List[traceback.FrameSummary],
) -> List[traceback.FrameSummary]:
    """
    Filters out commonly repeated frames of the asyncio internals from the given stack.
    """
    # strip the task factory frame in the vanilla event loop
    if (
        stack[-1].filename.endswith("asyncio/base_events.py")
        and stack[-1].name == "create_task"
    ):
        stack = stack[:-1]
    # strip the loop.create_task frame
    if (
        stack[-1].filename.endswith("asyncio/tasks.py")
        and stack[-1].name == "create_task"
    ):
        stack = stack[:-1]
    _cut_idx = 0
    for _cut_idx, f in reversed(list(enumerate(stack))):
        # uvloop
        if f.filename.endswith("asyncio/runners.py") and f.name == "run":
            break
        # vanilla
        if f.filename.endswith("asyncio/events.py") and f.name == "_run":
            break
    return stack[_cut_idx + 1 :]


def _extract_stack_from_task(
    task: "asyncio.Task[Any]",
) -> List[traceback.FrameSummary]:
    """
    Extracts the stack as a list of FrameSummary objects from an asyncio task.
    """
    frames: List[Any] = []
    coro = task._coro  # type: ignore
    while coro:
        f = getattr(coro, "cr_frame", getattr(coro, "gi_frame", None))
        if f is not None:
            frames.append(f)
        coro = getattr(coro, "cr_await", getattr(coro, "gi_yieldfrom", None))
    extracted_list = []
    checked: Set[str] = set()
    for f in frames:
        lineno = f.f_lineno
        co = f.f_code
        filename = co.co_filename
        name = co.co_name
        if filename not in checked:
            checked.add(filename)
            linecache.checkcache(filename)
        line = linecache.getline(filename, lineno, f.f_globals)
        extracted_list.append(traceback.FrameSummary(filename, lineno, name, line=line))
    return extracted_list


def _extract_stack_from_frame(frame: FrameType) -> List[traceback.FrameSummary]:
    stack = traceback.StackSummary.extract(traceback.walk_stack(frame))
    stack.reverse()
    return stack


def _extract_stack_from_exception(e: BaseException) -> List[traceback.FrameSummary]:
    stack = traceback.StackSummary.extract(traceback.walk_tb(e.__traceback__))
    stack.reverse()
    return stack


def get_default_args(func):
    signature = inspect.signature(func)
    return {
        k: v.default
        for k, v in signature.parameters.items()
        if v.default is not inspect.Parameter.empty
    }



from __future__ import annotations

import asyncio
from typing import Any, Dict, Optional

import aioconsole
from prompt_toolkit import PromptSession
from prompt_toolkit.input.base import Input
from prompt_toolkit.output.base import Output


class ConsoleProxy:
    # Runs inside the monitored event loop

    def __init__(
        self,
        stdin: Input,
        stdout: Output,
        host: str,
        port: int,
    ) -> None:
        self._host = host
        self._port = port
        self._stdin = stdin
        self._stdout = stdout

    async def __aenter__(self) -> ConsoleProxy:
        self._stdout.write_raw("\r\n")
        self._conn_reader, self._conn_writer = await asyncio.open_connection(
            self._host, self._port
        )
        self._closed = asyncio.Event()
        self._recv_task = asyncio.create_task(self._handle_received())
        self._input_task = asyncio.create_task(self._handle_user_input())
        return self

    async def __aexit__(self, *exc_info) -> Optional[bool]:
        self._input_task.cancel()
        await self._input_task
        self._conn_writer.close()
        try:
            await self._conn_writer.wait_closed()
        except (NotImplementedError, ConnectionResetError):
            pass
        self._conn_reader.feed_eof()
        await self._recv_task
        return None

    async def interact(self) -> None:
        try:
            await self._closed.wait()
        finally:
            self._closed.set()
            if not self._conn_writer.is_closing():
                self._conn_writer.write_eof()

    async def _handle_user_input(self) -> None:
        prompt_session: PromptSession[str] = PromptSession(
            input=self._stdin, output=self._stdout
        )
        try:
            while not self._conn_reader.at_eof():
                try:
                    user_input = await prompt_session.prompt_async("")
                    self._conn_writer.write(user_input.encode("utf8"))
                    self._conn_writer.write(b"\n")
                    await self._conn_writer.drain()
                except KeyboardInterrupt:
                    # Send Ctrl+C to the console server.
                    self._conn_writer.write(b"\x03")
                    await self._conn_writer.drain()
                except EOFError:
                    return
        except asyncio.CancelledError:
            pass
        finally:
            self._closed.set()

    async def _handle_received(self) -> None:
        try:
            while True:
                buf = await self._conn_reader.read(1024)
                if not buf:
                    return
                self._stdout.write_raw(buf.decode("utf8"))
                self._stdout.flush()
        except (ConnectionResetError, asyncio.CancelledError):
            pass
        finally:
            self._closed.set()


async def start(
    host: str,
    port: int,
    locals: Optional[Dict[str, Any]],
    monitor_loop: asyncio.AbstractEventLoop,
) -> asyncio.AbstractServer:
    async def _start() -> asyncio.AbstractServer:
        # Runs inside the monitored event loop
        def _factory(streams: Any = None) -> aioconsole.AsynchronousConsole:
            return aioconsole.AsynchronousConsole(
                locals=locals, streams=streams, loop=monitor_loop
            )

        server = await aioconsole.start_interactive_server(
            host=host,
            port=port,
            factory=_factory,
            loop=monitor_loop,
        )
        return server

    console_future = asyncio.wrap_future(
        asyncio.run_coroutine_threadsafe(
            _start(),
            loop=monitor_loop,
        )
    )
    return await console_future


async def close(
    server: asyncio.AbstractServer,
    monitor_loop: asyncio.AbstractEventLoop,
) -> None:
    async def _close() -> None:
        # Runs inside the monitored event loop
        try:
            server.close()
            await server.wait_closed()
        except NotImplementedError:
            pass

    close_future = asyncio.wrap_future(
        asyncio.run_coroutine_threadsafe(
            _close(),
            loop=monitor_loop,
        )
    )
    await close_future


async def proxy(sin: Input, sout: Output, host: str, port: int) -> None:
    async with ConsoleProxy(sin, sout, host, port) as proxy:
        await proxy.interact()



import asyncio

import aiomonitor
from aiomonitor.task import preserve_termination_log


@preserve_termination_log
async def should_have_run_until_exit():
    print("should_have_run_until_exit: begin")
    await asyncio.sleep(5)
    print("should_have_run_until_exit: cancel")
    raise asyncio.CancelledError("cancelled-suspiciously")


async def chain3():
    await asyncio.sleep(10)


async def chain2():
    t = asyncio.create_task(chain3())
    try:
        await asyncio.sleep(10)
    finally:
        t.cancel()
        await t


async def chain1():
    t = asyncio.create_task(chain2())
    try:
        await asyncio.sleep(10)
    finally:
        t.cancel()
        await t


async def chain_main():
    try:
        while True:
            t = asyncio.create_task(chain1())
            await asyncio.sleep(0.5)
            t.cancel()
    finally:
        print("terminating chain_main")


async def self_cancel_main():
    async def do_self_cancel(tick):
        await asyncio.sleep(tick)
        raise asyncio.CancelledError("self-cancelled")

    try:
        while True:
            await asyncio.sleep(0.21)
            asyncio.create_task(do_self_cancel(0.2))
    finally:
        print("terminating self_cancel_main")


async def unhandled_exc_main():
    async def do_unhandled(tick):
        await asyncio.sleep(tick)
        return 1 / 0

    try:
        while True:
            try:
                await asyncio.create_task(do_unhandled(0.2))
            except ZeroDivisionError:
                continue
    finally:
        print("terminating unhandled_exc_main")


async def main():
    loop = asyncio.get_running_loop()
    with aiomonitor.start_monitor(
        loop,
        hook_task_factory=True,
        max_termination_history=10,
    ):
        asyncio.create_task(should_have_run_until_exit())
        chain_main_task = asyncio.create_task(chain_main())
        self_cancel_main_task = asyncio.create_task(self_cancel_main())
        unhandled_exc_main_task = asyncio.create_task(unhandled_exc_main())
        try:
            print("running some test workloads for 10 seconds ...")
            await asyncio.sleep(10)
        finally:
            print("stopping ...")
            chain_main_task.cancel()
            self_cancel_main_task.cancel()
            unhandled_exc_main_task.cancel()
            await asyncio.gather(
                chain_main_task,
                self_cancel_main_task,
                unhandled_exc_main_task,
                return_exceptions=True,
            )
        print("waiting inspection ...")
        while True:
            await asyncio.sleep(10)


if __name__ == "__main__":
    try:
        asyncio.run(main())
    except KeyboardInterrupt:
        print("terminated")



from __future__ import annotations

import sys
import traceback
from dataclasses import dataclass
from typing import List, NamedTuple, Optional

if sys.version_info >= (3, 11):
    from enum import StrEnum
else:
    from backports.strenum import StrEnum


@dataclass
class FormattedLiveTaskInfo:
    task_id: str
    state: str
    name: str
    coro: str
    created_location: str
    since: str


@dataclass
class FormattedTerminatedTaskInfo:
    task_id: str
    name: str
    coro: str
    started_since: str
    terminated_since: str


class FormatItemTypes(StrEnum):
    HEADER = "header"
    CONTENT = "content"


class FormattedStackItem(NamedTuple):
    type: FormatItemTypes
    content: str


@dataclass
class TerminatedTaskInfo:
    id: str
    name: str
    coro: str
    started_at: float
    terminated_at: float
    cancelled: bool
    termination_stack: Optional[List[traceback.FrameSummary]] = None
    canceller_stack: Optional[List[traceback.FrameSummary]] = None
    exc_repr: Optional[str] = None
    persistent: bool = False


@dataclass
class CancellationChain:
    target_id: str
    canceller_id: str
    canceller_stack: Optional[List[traceback.FrameSummary]] = None



from __future__ import annotations

import asyncio
import contextlib
import contextvars
import functools
import logging
import sys
import textwrap
import threading
import time
import traceback
import weakref
from asyncio.coroutines import _format_coroutine  # type: ignore
from datetime import timedelta
from types import TracebackType
from typing import (
    Any,
    Awaitable,
    Coroutine,
    Dict,
    Final,
    Generator,
    List,
    Optional,
    Sequence,
    Type,
    TypeVar,
    cast,
)

import janus
from aiohttp import web
from prompt_toolkit.contrib.telnet.server import TelnetServer

from .exceptions import MissingTask
from .task import TracedTask, persistent_coro
from .termui.commands import interact
from .types import (
    CancellationChain,
    FormatItemTypes,
    FormattedLiveTaskInfo,
    FormattedStackItem,
    FormattedTerminatedTaskInfo,
    TerminatedTaskInfo,
)
from .utils import (
    _extract_stack_from_exception,
    _extract_stack_from_frame,
    _extract_stack_from_task,
    _filter_stack,
    _format_filename,
    _format_task,
    _format_terminated_task,
    _format_timedelta,
    get_default_args,
)
from .webui.app import init_webui

__all__ = (
    "Monitor",
    "start_monitor",
)

log = logging.getLogger(__name__)

MONITOR_HOST: Final = "127.0.0.1"
MONITOR_TERMUI_PORT: Final = 20101
MONITOR_WEBUI_PORT: Final = 20102
CONSOLE_PORT: Final = 20103

T = TypeVar("T")
T_co = TypeVar("T_co", covariant=True)


def task_by_id(
    taskid: int, loop: asyncio.AbstractEventLoop
) -> "Optional[asyncio.Task[Any]]":
    tasks = asyncio.all_tasks(loop=loop)
    return next(filter(lambda t: id(t) == taskid, tasks), None)


async def cancel_task(task: "asyncio.Task[Any]") -> None:
    with contextlib.suppress(asyncio.CancelledError):
        task.cancel()
        await task


class Monitor:
    prompt: str
    """
    The string that prompts you to enter a command, defaults to ``"monitor >>> "``
    """

    _event_loop_thread_id: Optional[int] = None

    console_locals: Dict[str, Any]
    _termui_tasks: weakref.WeakSet[asyncio.Task[Any]]

    _created_traceback_chains: weakref.WeakKeyDictionary[
        asyncio.Task[Any],
        weakref.ReferenceType[asyncio.Task[Any]],
    ]
    _created_tracebacks: weakref.WeakKeyDictionary[
        asyncio.Task[Any], List[traceback.FrameSummary]
    ]
    _terminated_tasks: Dict[str, TerminatedTaskInfo]
    _terminated_history: List[str]
    _termination_info_queue: janus.Queue[TerminatedTaskInfo]
    _canceller_chain: Dict[str, str]
    _canceller_stacks: Dict[str, List[traceback.FrameSummary] | None]
    _cancellation_chain_queue: janus.Queue[CancellationChain]

    def __init__(
        self,
        loop: asyncio.AbstractEventLoop,
        *,
        host: str = MONITOR_HOST,
        termui_port: int = MONITOR_TERMUI_PORT,
        webui_port: int = MONITOR_WEBUI_PORT,
        console_port: int = CONSOLE_PORT,
        console_enabled: bool = True,
        hook_task_factory: bool = False,
        max_termination_history: int = 1000,
        locals: Optional[Dict[str, Any]] = None,
    ) -> None:
        self._monitored_loop = loop or asyncio.get_running_loop()
        self._host = host
        self._termui_port = termui_port
        self._webui_port = webui_port
        self._console_port = console_port
        self._console_enabled = console_enabled
        if locals is None:
            self.console_locals = {"__name__": "__console__", "__doc__": None}
        else:
            self.console_locals = locals

        self.prompt = "monitor >>> "
        log.info(
            "Starting aiomonitor at telnet://%(host)s:%(tport)d and http://%(host)s:%(wport)d",
            {
                "host": host,
                "tport": termui_port,
                "wport": webui_port,
            },
        )

        self._closed = False
        self._started = False
        self._termui_tasks = weakref.WeakSet()

        self._hook_task_factory = hook_task_factory
        self._created_traceback_chains = weakref.WeakKeyDictionary()
        self._created_tracebacks = weakref.WeakKeyDictionary()
        self._terminated_tasks = {}
        self._canceller_chain = {}
        self._canceller_stacks = {}
        self._terminated_history = []
        self._max_termination_history = max_termination_history

        self._ui_started = threading.Event()
        self._ui_thread = threading.Thread(target=self._ui_main, args=(), daemon=True)

    @property
    def host(self) -> str:
        """
        The current hostname to bind the monitor server.
        """
        return self._host

    @property
    def port(self) -> int:
        """
        The port number to bind the monitor server.
        """
        return self._termui_port

    def __repr__(self) -> str:
        name = self.__class__.__name__
        return "<{name}: {host}:{port}>".format(
            name=name, host=self._host, port=self._termui_port
        )

    def start(self) -> None:
        """
        Starts monitoring thread, where telnet server is executed.
        """
        assert not self._closed
        assert not self._started
        self._started = True
        self._original_task_factory = self._monitored_loop.get_task_factory()
        if self._hook_task_factory:
            self._monitored_loop.set_task_factory(self._create_task)
        self._event_loop_thread_id = threading.get_ident()
        self._ui_thread.start()
        self._ui_started.wait()

    @property
    def closed(self) -> bool:
        """
        A flag indicates if monitor was closed, currntly instance of
        :class:`Monitor` can not be reused. For new monitor, new instance
        should be created.
        """
        return self._closed

    def __enter__(self) -> Monitor:
        if not self._started:
            self.start()
        return self

    # exc_type should be Optional[Type[BaseException]], but
    # this runs into https://github.com/python/typing/issues/266
    # on Python 3.5.
    def __exit__(
        self,
        exc_type: Any,
        exc_value: Optional[BaseException],
        traceback: Optional[TracebackType],
    ) -> None:
        self.close()

    def close(self) -> None:
        """
        Joins background thread, and cleans up resources.
        """
        assert self._started, "The monitor must have been started to close it."
        if not self._closed:
            self._ui_loop.call_soon_threadsafe(
                self._ui_forever_future.cancel,
            )
            self._monitored_loop.set_task_factory(self._original_task_factory)
            self._ui_thread.join()
            self._closed = True

    def format_running_task_list(
        self, filter_: str, persistent: bool
    ) -> Sequence[FormattedLiveTaskInfo]:
        all_running_tasks = asyncio.all_tasks(loop=self._monitored_loop)
        tasks = []
        for task in sorted(all_running_tasks, key=id):
            taskid = str(id(task))
            if isinstance(task, TracedTask):
                coro_repr = _format_coroutine(task._orig_coro).partition(" ")[0]
                if persistent and task._orig_coro not in persistent_coro:
                    continue
            else:
                coro_repr = _format_coroutine(task.get_coro()).partition(" ")[0]
                if persistent:
                    # untracked tasks should be skipped when showing persistent ones only
                    continue
            if filter_ and (
                filter_ not in coro_repr and filter_ not in task.get_name()
            ):
                continue
            creation_stack = self._created_tracebacks.get(task)
            # Some values are masked as "-" when they are unavailable
            # if it's the root task/coro or if the task factory is not applied.
            if not creation_stack:
                created_location = "-"
            else:
                creation_stack = _filter_stack(creation_stack)
                fn = _format_filename(creation_stack[-1].filename)
                lineno = creation_stack[-1].lineno
                created_location = f"{fn}:{lineno}"
            if isinstance(task, TracedTask):
                running_since = _format_timedelta(
                    timedelta(
                        seconds=(time.perf_counter() - task._started_at),
                    )
                )
            else:
                running_since = "-"
            tasks.append(
                FormattedLiveTaskInfo(
                    taskid,
                    task._state,
                    task.get_name(),
                    coro_repr,
                    created_location,
                    running_since,
                )
            )
        return tasks

    def format_terminated_task_list(
        self, filter_: str, persistent: bool
    ) -> Sequence[FormattedTerminatedTaskInfo]:
        terminated_tasks = self._terminated_tasks.values()
        tasks = []
        for item in sorted(
            terminated_tasks,
            key=lambda info: info.terminated_at,
            reverse=True,
        ):
            if persistent and not item.persistent:
                continue
            if filter_ and (filter_ not in item.coro and filter_ not in item.name):
                continue
            started_since = _format_timedelta(
                timedelta(seconds=time.perf_counter() - item.started_at)
            )
            terminated_since = _format_timedelta(
                timedelta(seconds=time.perf_counter() - item.terminated_at)
            )
            tasks.append(
                FormattedTerminatedTaskInfo(
                    str(item.id),
                    item.name,
                    item.coro,
                    started_since,
                    terminated_since,
                )
            )
        return tasks

    async def cancel_monitored_task(self, task_id: str | int) -> str:
        task_id_ = int(task_id)
        task = task_by_id(task_id_, self._monitored_loop)
        if task is not None:
            if self._monitored_loop == asyncio.get_running_loop():
                await cancel_task(task)
            else:
                fut = asyncio.wrap_future(
                    asyncio.run_coroutine_threadsafe(
                        cancel_task(task), loop=self._monitored_loop
                    )
                )
                await fut
            if isinstance(task, TracedTask):
                coro_repr = _format_coroutine(task._orig_coro).partition(" ")[0]
            else:
                coro_repr = _format_coroutine(task.get_coro()).partition(" ")[0]
            return coro_repr
        else:
            raise ValueError("Invalid or non-existent task ID", task_id)

    def format_running_task_stack(
        self,
        task_id: str | int,
    ) -> Sequence[FormattedStackItem]:
        depth = 0
        task_id_ = int(task_id)
        task = task_by_id(task_id_, self._monitored_loop)
        if task is None:
            raise MissingTask(task_id_)
        task_chain: List[asyncio.Task[Any]] = []
        while task is not None:
            task_chain.append(task)
            task_ref = self._created_traceback_chains.get(task)
            task = task_ref() if task_ref is not None else None
        prev_task = None
        formatted_stack_list = []
        for task in reversed(task_chain):
            if depth == 0:
                formatted_stack_list.append(
                    FormattedStackItem(
                        FormatItemTypes.HEADER,
                        (
                            "Stack of the root task or coroutine scheduled "
                            "in the event loop (most recent call last)"
                        ),
                    )
                )
            elif depth > 0:
                assert prev_task is not None
                formatted_stack_list.append(
                    FormattedStackItem(
                        FormatItemTypes.HEADER,
                        (
                            "Stack of %s when creating the next task "
                            "(most recent call last)" % _format_task(prev_task)
                        ),
                    )
                )
            stack = self._created_tracebacks.get(task)
            if stack is None:
                formatted_stack_list.append(
                    FormattedStackItem(
                        FormatItemTypes.CONTENT,
                        (
                            "No stack available (maybe it is a native code, "
                            "a synchronous callback function, "
                            "or the event loop itself)"
                        ),
                    )
                )
            else:
                stack = _filter_stack(stack)
                formatted_stack_list.append(
                    FormattedStackItem(
                        FormatItemTypes.CONTENT,
                        textwrap.dedent("".join(traceback.format_list(stack))),
                    )
                )
            prev_task = task
            depth += 1
        task = task_chain[0]
        formatted_stack_list.append(
            FormattedStackItem(
                FormatItemTypes.HEADER,
                "Stack of %s (most recent call last)" % _format_task(task),
            )
        )
        stack = _extract_stack_from_task(task)
        if not stack:
            formatted_stack_list.append(
                FormattedStackItem(
                    FormatItemTypes.CONTENT,
                    "No stack available for %s" % _format_task(task),
                )
            )
        else:
            formatted_stack_list.append(
                FormattedStackItem(
                    FormatItemTypes.CONTENT,
                    textwrap.dedent("".join(traceback.format_list(stack))),
                )
            )
        return formatted_stack_list

    def format_terminated_task_stack(
        self,
        trace_id: str,
    ) -> Sequence[FormattedStackItem]:
        depth = 0
        tinfo_chain: List[TerminatedTaskInfo] = []
        while trace_id is not None:
            tinfo_chain.append(self._terminated_tasks[trace_id])
            trace_id = self._canceller_chain.get(trace_id)  # type: ignore
        prev_tinfo = None
        formatted_stack_list = []
        for tinfo in reversed(tinfo_chain):
            if depth == 0:
                formatted_stack_list.append(
                    FormattedStackItem(
                        FormatItemTypes.HEADER,
                        (
                            "Stack of the root task or coroutine "
                            "scheduled in the event loop"
                            "(most recent call last)"
                        ),
                    )
                )
            elif depth > 0:
                assert prev_tinfo is not None
                formatted_stack_list.append(
                    FormattedStackItem(
                        FormatItemTypes.HEADER,
                        (
                            "Stack of %s when creating the next task "
                            "(most recent call last)"
                            % _format_terminated_task(prev_tinfo)
                        ),
                    )
                )
            stack = tinfo.canceller_stack
            if stack is None:
                formatted_stack_list.append(
                    FormattedStackItem(
                        FormatItemTypes.CONTENT,
                        (
                            "No stack available "
                            "(maybe it is a self-raised cancellation or exception)"
                        ),
                    )
                )
            else:
                stack = _filter_stack(stack)
                formatted_stack_list.append(
                    FormattedStackItem(
                        FormatItemTypes.CONTENT,
                        textwrap.dedent("".join(traceback.format_list(stack))),
                    )
                )
            prev_tinfo = tinfo
            depth += 1
        tinfo = tinfo_chain[0]
        formatted_stack_list.append(
            FormattedStackItem(
                FormatItemTypes.HEADER,
                "Stack of %s (most recent call last)" % _format_terminated_task(tinfo),
            )
        )
        stack = tinfo.termination_stack
        if not stack:
            formatted_stack_list.append(
                FormattedStackItem(
                    FormatItemTypes.CONTENT,
                    (
                        "No stack available for %s (the task has run to completion)"
                        % _format_terminated_task(tinfo)
                    ),
                )
            )
        else:
            formatted_stack_list.append(
                FormattedStackItem(
                    FormatItemTypes.CONTENT,
                    textwrap.dedent("".join(traceback.format_list(stack))),
                )
            )
        return formatted_stack_list

    async def _coro_wrapper(self, coro: Awaitable[T_co]) -> T_co:
        myself = asyncio.current_task()
        assert isinstance(myself, TracedTask)
        try:
            return await coro
        except BaseException as e:
            myself._termination_stack = _extract_stack_from_exception(e)[:-1]
            raise

    def _create_task(
        self,
        loop: asyncio.AbstractEventLoop,
        coro: Coroutine[Any, Any, T_co] | Generator[Any, None, T_co],
        *,
        name: str | None = None,
        context: contextvars.Context | None = None,
    ) -> asyncio.Future[T_co]:
        assert loop is self._monitored_loop
        try:
            parent_task = asyncio.current_task()
        except RuntimeError:
            parent_task = None
        persistent = coro in persistent_coro
        task = TracedTask(
            self._coro_wrapper(coro),  # type: ignore
            termination_info_queue=self._termination_info_queue.sync_q,
            cancellation_chain_queue=self._cancellation_chain_queue.sync_q,
            persistent=persistent,
            loop=self._monitored_loop,
            name=name,  # since Python 3.8
            context=context,  # since Python 3.11
        )
        task._orig_coro = cast(Coroutine[Any, Any, T_co], coro)
        self._created_tracebacks[task] = _extract_stack_from_frame(sys._getframe())[
            :-1
        ]  # strip this wrapper method
        if parent_task is not None:
            self._created_traceback_chains[task] = weakref.ref(parent_task)
        return task

    def _ui_main(self) -> None:
        asyncio.run(self._ui_main_async())

    async def _ui_main_async(self) -> None:
        loop = asyncio.get_running_loop()
        self._termination_info_queue = janus.Queue()
        self._cancellation_chain_queue = janus.Queue()
        self._ui_loop = loop
        self._ui_forever_future = loop.create_future()
        self._ui_termination_handler_task = loop.create_task(
            self._ui_handle_termination_updates()
        )
        self._ui_cancellation_handler_task = loop.create_task(
            self._ui_handle_cancellation_updates()
        )
        telnet_server = TelnetServer(
            interact=functools.partial(interact, self),
            host=self._host,
            port=self._termui_port,
        )
        webui_app = await init_webui(self)
        webui_runner = web.AppRunner(webui_app)
        await webui_runner.setup()
        webui_site = web.TCPSite(
            webui_runner,
            str(self._host),
            self._webui_port,
            reuse_port=True,
        )
        await webui_site.start()
        telnet_server.start()
        await asyncio.sleep(0)
        self._ui_started.set()
        try:
            await self._ui_forever_future
        except asyncio.CancelledError:
            pass
        finally:
            termui_tasks = {*self._termui_tasks}
            for termui_task in termui_tasks:
                termui_task.cancel()
            await asyncio.gather(*termui_tasks, return_exceptions=True)
            self._ui_termination_handler_task.cancel()
            self._ui_cancellation_handler_task.cancel()
            with contextlib.suppress(asyncio.CancelledError):
                await self._ui_termination_handler_task
            with contextlib.suppress(asyncio.CancelledError):
                await self._ui_cancellation_handler_task
            await telnet_server.stop()
            await webui_runner.cleanup()

    async def _ui_handle_termination_updates(self) -> None:
        while True:
            try:
                update: TerminatedTaskInfo = (
                    await self._termination_info_queue.async_q.get()
                )
            except asyncio.CancelledError:
                return
            self._terminated_tasks[update.id] = update
            if not update.persistent:
                self._terminated_history.append(update.id)
            # canceller stack is already put in _ui_handle_cancellation_updates()
            if canceller_stack := self._canceller_stacks.pop(update.id, None):
                update.canceller_stack = canceller_stack
            while len(self._terminated_history) > self._max_termination_history:
                removed_id = self._terminated_history.pop(0)
                self._terminated_tasks.pop(removed_id, None)
                self._canceller_chain.pop(removed_id, None)
                self._canceller_stacks.pop(removed_id, None)

    async def _ui_handle_cancellation_updates(self) -> None:
        while True:
            try:
                update: CancellationChain = (
                    await self._cancellation_chain_queue.async_q.get()
                )
            except asyncio.CancelledError:
                return
            self._canceller_stacks[update.target_id] = update.canceller_stack
            self._canceller_chain[update.target_id] = update.canceller_id


def start_monitor(
    loop: asyncio.AbstractEventLoop,
    *,
    monitor_cls: Type[Monitor] = Monitor,
    host: str = MONITOR_HOST,
    port: int = MONITOR_TERMUI_PORT,  # kept the name for backward compatibility
    console_port: int = CONSOLE_PORT,
    webui_port: int = MONITOR_WEBUI_PORT,
    console_enabled: bool = True,
    hook_task_factory: bool = False,
    max_termination_history: Optional[int] = None,
    locals: Optional[Dict[str, Any]] = None,
) -> Monitor:
    """
    Factory function, creates instance of :class:`Monitor` and starts
    monitoring thread.

    :param Type[Monitor] monitor: Monitor class to use
    :param str host: hostname to serve monitor telnet server
    :param int port: monitor port (terminal UI), by default 20101
    :param int webui_port: monitor port (web UI), by default 20102
    :param int console_port: python REPL port, by default 20103
    :param bool console_enabled: flag indicates if python REPL is requred
        to start with instance of monitor.
    :param dict locals: dictionary with variables exposed in python console
        environment
    """
    m = monitor_cls(
        loop,
        host=host,
        termui_port=port,
        webui_port=webui_port,
        console_port=console_port,
        console_enabled=console_enabled,
        hook_task_factory=hook_task_factory,
        max_termination_history=(
            max_termination_history
            if max_termination_history is not None
            else get_default_args(monitor_cls.__init__)["max_termination_history"]
        ),
        locals=locals,
    )
    m.start()
    return m



"""Debugging monitor for asyncio app with asynchronous python REPL capabilities

To enable the monitor, just use context manager protocol with start function::

    import asyncio
    import aiomonitor

    loop = asyncio.get_event_loop()
    with aiomonitor.start_monitor():
        print("Now you can connect with: nc localhost 20101")
        loop.run_forever()

Alternatively you can use more verbose try/finally approach::

    m = Monitor()
    m.start()
    try:
        loop.run_forever()
    finally:
        m.close()
"""

from importlib.metadata import version

from .monitor import (
    CONSOLE_PORT,
    MONITOR_HOST,
    MONITOR_TERMUI_PORT,
    MONITOR_WEBUI_PORT,
    Monitor,
    start_monitor,
)
from .termui.commands import monitor_cli

MONITOR_PORT = MONITOR_TERMUI_PORT  # for backward compatibility

__all__ = (
    "Monitor",
    "start_monitor",
    "monitor_cli",
    "MONITOR_HOST",
    "MONITOR_PORT",
    "MONITOR_TERMUI_PORT",
    "MONITOR_WEBUI_PORT",
    "CONSOLE_PORT",
)
__version__ = version("aiomonitor")



import asyncio
import logging

import uvloop
from aiohttp import web

import aiomonitor


async def inner2() -> None:
    await asyncio.sleep(100)


async def inner1(tg: asyncio.TaskGroup) -> None:
    t = tg.create_task(inner2())
    await t


async def simple(request: web.Request) -> web.Response:
    print("Start sleeping")
    async with asyncio.TaskGroup() as tg:
        tg.create_task(inner1(tg))
    print("Finished sleeping")
    return web.Response(text="Simple answer")


async def main() -> None:
    loop = asyncio.get_running_loop()
    app = web.Application()
    app.router.add_get("/simple", simple)
    with aiomonitor.start_monitor(loop, hook_task_factory=True):
        await web._run_app(app, port=8090, host="localhost")


if __name__ == "__main__":
    logging.basicConfig()
    logging.getLogger("aiomonitor").setLevel(logging.DEBUG)
    uvloop.install()
    try:
        asyncio.run(main())
    except KeyboardInterrupt:
        pass



from __future__ import annotations

import asyncio
import collections
import fcntl
import os
import struct
import sys
import telnetlib
import termios
from typing import Dict, Optional, TextIO, Tuple

ModeDef = collections.namedtuple(
    "ModeDef", ["iflag", "oflag", "cflag", "lflag", "ispeed", "ospeed", "cc"]
)


class TelnetClient:
    """
    A minimal telnet client-side protocol implementation to support `prompt_toolkit`.
    Some details like terminal mode handling is taken from telnetlib3.
    (https://github.com/jquast/telnetlib3)
    """

    def __init__(
        self,
        host: str,
        port: int,
        stdin: Optional[TextIO] = None,
        stdout: Optional[TextIO] = None,
    ) -> None:
        self._host = host
        self._port = port
        self._term = os.environ.get("TERM", "unknown")
        self._stdin = stdin or sys.stdin
        self._stdout = stdout or sys.stdout
        try:
            self._isatty = os.path.sameopenfile(
                self._stdin.fileno(), self._stdout.fileno()
            )
        except (NotImplementedError, ValueError):
            self._isatty = False
        self._remote_options: Dict[bytes, bool] = collections.defaultdict(lambda: False)

    def get_mode(self) -> Optional[ModeDef]:
        if self._isatty:
            try:
                return ModeDef(*termios.tcgetattr(self._stdin.fileno()))
            except termios.error:
                return None
        return None

    def set_mode(self, mode: ModeDef) -> None:
        termios.tcsetattr(sys.stdin.fileno(), termios.TCSAFLUSH, list(mode))

    def restore_mode(self) -> None:
        if self._isatty and self._saved_mode is not None:
            termios.tcsetattr(
                self._stdin.fileno(), termios.TCSAFLUSH, list(self._saved_mode)
            )

    def determine_mode(self, mode: ModeDef) -> ModeDef:
        # Reference: https://github.com/jquast/telnetlib3/blob/b90616f/telnetlib3/client_shell.py#L76
        if not self._remote_options[telnetlib.ECHO]:
            return mode
        iflag = mode.iflag & ~(
            termios.BRKINT
            | termios.ICRNL  # Do not send INTR signal on break
            | termios.INPCK  # Do not map CR to NL on input
            | termios.ISTRIP  # Disable input parity checking
            | termios.IXON  # Do not strip input characters to 7 bits
        )
        cflag = mode.cflag & ~(termios.CSIZE | termios.PARENB)
        cflag = cflag | termios.CS8
        lflag = mode.lflag & ~(
            termios.ICANON | termios.IEXTEN | termios.ISIG | termios.ECHO
        )
        oflag = mode.oflag & ~(termios.OPOST | termios.ONLCR)
        cc = list(mode.cc)
        cc[termios.VMIN] = 1
        cc[termios.VTIME] = 0
        return ModeDef(
            iflag=iflag,
            oflag=oflag,
            cflag=cflag,
            lflag=lflag,
            ispeed=mode.ispeed,
            ospeed=mode.ospeed,
            cc=cc,
        )

    async def _create_stdio_streams(
        self,
    ) -> Tuple[asyncio.StreamReader, asyncio.StreamWriter]:
        loop = asyncio.get_running_loop()
        stdin_reader = asyncio.StreamReader()
        stdin_protocol = asyncio.StreamReaderProtocol(stdin_reader)
        await loop.connect_read_pipe(lambda: stdin_protocol, self._stdin)
        write_target = self._stdin if self._isatty else self._stdout
        writer_transport, writer_protocol = await loop.connect_write_pipe(
            asyncio.streams.FlowControlMixin,
            write_target,
        )
        stdout_writer = asyncio.StreamWriter(
            writer_transport, writer_protocol, stdin_reader, loop
        )
        return stdin_reader, stdout_writer

    async def __aenter__(self) -> TelnetClient:
        self._conn_reader, self._conn_writer = await asyncio.open_connection(
            self._host, self._port
        )
        self._closed = asyncio.Event()
        self._stdin_reader, self._stdout_writer = await self._create_stdio_streams()
        self._recv_task = asyncio.create_task(self._handle_received())
        self._input_task = asyncio.create_task(self._handle_user_input())
        await asyncio.sleep(0.3)  # wait for negotiation to complete
        self._saved_mode = self.get_mode()
        if self._isatty:
            assert self._saved_mode is not None
            self.set_mode(self.determine_mode(self._saved_mode))
        return self

    async def __aexit__(self, *exc_info) -> Optional[bool]:
        try:
            self._input_task.cancel()
            await self._input_task
            self._conn_writer.close()
            try:
                await self._conn_writer.wait_closed()
            except NotImplementedError:
                pass
            self._conn_reader.feed_eof()
            await self._recv_task
            self._stdout_writer.close()
            try:
                await self._stdout_writer.wait_closed()
            except NotImplementedError:
                pass
        finally:
            self.restore_mode()
        return None

    async def interact(self) -> None:
        try:
            await self._closed.wait()
        finally:
            self._closed.set()
            self._conn_writer.write_eof()

    async def _handle_user_input(self) -> None:
        try:
            while True:
                buf = await self._stdin_reader.read(128)
                if not buf:
                    return
                self._conn_writer.write(buf)
                await self._conn_writer.drain()
        except asyncio.CancelledError:
            pass

    async def _handle_nego(self, command: bytes, option: bytes) -> None:
        if command == telnetlib.DO and option == telnetlib.TTYPE:
            self._conn_writer.write(telnetlib.IAC + telnetlib.WILL + telnetlib.TTYPE)
            self._conn_writer.write(telnetlib.IAC + telnetlib.SB + telnetlib.TTYPE)
            self._conn_writer.write(telnetlib.BINARY + self._term.encode("ascii"))
            self._conn_writer.write(telnetlib.IAC + telnetlib.SE)
            await self._conn_writer.drain()
        elif command == telnetlib.DO and option == telnetlib.NAWS:
            self._conn_writer.write(telnetlib.IAC + telnetlib.WILL + telnetlib.NAWS)
            self._conn_writer.write(telnetlib.IAC + telnetlib.SB + telnetlib.NAWS)
            fmt = "HHHH"
            buf = b"\x00" * struct.calcsize(fmt)
            try:
                buf = fcntl.ioctl(self._stdin.fileno(), termios.TIOCGWINSZ, buf)
                rows, cols, _, _ = struct.unpack(fmt, buf)
                self._conn_writer.write(struct.pack(">HH", cols, rows))
                self._conn_writer.write(telnetlib.IAC + telnetlib.SE)
            except OSError:
                rows, cols = 22, 80
            await self._conn_writer.drain()
        elif command == telnetlib.WILL:
            self._remote_options[option] = True
            if option in (telnetlib.ECHO, telnetlib.BINARY, telnetlib.SGA):
                self._conn_writer.write(telnetlib.IAC + telnetlib.DO + option)
                await self._conn_writer.drain()
        elif command == telnetlib.WONT:
            self._remote_options[option] = False

    async def _handle_sb(self, option: bytes, chunk: bytes) -> None:
        pass

    async def _handle_received(self):
        buf = b""
        try:
            while not self._conn_reader.at_eof():
                buf += await self._conn_reader.read(128)
                while buf:
                    cmd_begin = buf.find(telnetlib.IAC)
                    if cmd_begin == -1:
                        self._stdout_writer.write(buf)
                        await self._stdout_writer.drain()
                        buf = b""
                    else:
                        if cmd_begin >= len(buf) - 2:
                            buf += await self._conn_reader.readexactly(
                                3 - (len(buf) - cmd_begin),
                            )
                        command = buf[cmd_begin + 1 : cmd_begin + 2]
                        option = buf[cmd_begin + 2 : cmd_begin + 3]
                        buf_before, buf_after = buf[:cmd_begin], buf[cmd_begin + 3 :]
                        self._stdout_writer.write(buf_before)
                        await self._stdout_writer.drain()
                        if command in (
                            telnetlib.WILL,
                            telnetlib.WONT,
                            telnetlib.DO,
                            telnetlib.DONT,
                        ):
                            await self._handle_nego(command, option)
                        elif command == telnetlib.SB:
                            subnego_end = buf_after.find(telnetlib.IAC + telnetlib.SE)
                            if subnego_end == -1:
                                subnego_chunk = (
                                    buf_after
                                    + (
                                        await self._conn_reader.readuntil(
                                            telnetlib.IAC + telnetlib.SE
                                        )
                                    )[:-2]
                                )
                                buf_after = b""
                            else:
                                subnego_chunk = buf_after[:subnego_end]
                                buf_after = buf_after[subnego_end + 2 :]
                            await self._handle_sb(option, subnego_chunk)
                        buf = buf_after
        finally:
            self._closed.set()



from __future__ import annotations

import asyncio
import dataclasses
import sys
from importlib.metadata import version
from pathlib import Path
from typing import TYPE_CHECKING, Dict, Mapping, Tuple

if sys.version_info >= (3, 11):
    from enum import StrEnum
else:
    from backports.strenum import StrEnum

import trafaret as t
from aiohttp import web
from jinja2 import Environment, PackageLoader, select_autoescape

from .utils import APIParams, check_params

if TYPE_CHECKING:
    from ..monitor import Monitor


@dataclasses.dataclass
class WebUIContext:
    monitor: Monitor
    jenv: Environment


class TaskTypes(StrEnum):
    RUNNING = "running"
    TERMINATED = "terminated"


@dataclasses.dataclass
class TaskTypeParams(APIParams):
    task_type: TaskTypes

    @classmethod
    def get_checker(cls):
        return t.Dict(
            {
                t.Key("task_type", default=TaskTypes.RUNNING): t.Enum(
                    TaskTypes.RUNNING,
                    TaskTypes.TERMINATED,
                ),
            }
        )


@dataclasses.dataclass
class TaskIdParams(APIParams):
    task_id: str

    @classmethod
    def get_checker(cls) -> t.Trafaret:
        return t.Dict(
            {
                t.Key("task_id"): t.String,
            }
        )


@dataclasses.dataclass
class ListFilterParams(APIParams):
    filter: str
    persistent: bool

    @classmethod
    def get_checker(cls) -> t.Trafaret:
        return t.Dict(
            {
                t.Key("filter", default=""): t.String(allow_blank=True),
                t.Key("persistent", default=False): t.ToBool,
            }
        )


@dataclasses.dataclass
class NavigationItem:
    title: str
    current: bool


nav_menus: Mapping[str, NavigationItem] = {
    "/": NavigationItem(
        title="Dashboard",
        current=False,
    ),
    "/about": NavigationItem(
        title="About",
        current=False,
    ),
}


def get_navigation_info(
    route: str,
) -> Tuple[NavigationItem, Mapping[str, NavigationItem]]:
    nav_items: Dict[str, NavigationItem] = {}
    current_item = None
    for path, item in nav_menus.items():
        is_current = path == route
        nav_items[path] = NavigationItem(item.title, is_current)
        if is_current:
            current_item = item
    if current_item is None:
        raise web.HTTPNotFound
    return current_item, nav_items


async def show_list_page(request: web.Request) -> web.Response:
    ctx: WebUIContext = request.app["ctx"]
    nav_info, nav_items = get_navigation_info(request.path)
    template = ctx.jenv.get_template("index.html")
    async with check_params(request, TaskTypeParams) as params:
        output = template.render(
            navigation=nav_items,
            page={
                "title": nav_info.title,
            },
            current_list_type=params.task_type,
            list_types=[
                {"id": TaskTypes.RUNNING, "title": "Running"},
                {"id": TaskTypes.TERMINATED, "title": "Terminated"},
            ],
        )
        return web.Response(body=output, content_type="text/html")


async def show_about_page(request: web.Request) -> web.Response:
    ctx: WebUIContext = request.app["ctx"]
    nav_info, nav_items = get_navigation_info(request.path)
    template = ctx.jenv.get_template("about.html")
    output = template.render(
        navigation=nav_items,
        page={
            "title": nav_info.title,
        },
    )
    return web.Response(body=output, content_type="text/html")


async def show_trace_page(request: web.Request) -> web.Response:
    ctx: WebUIContext = request.app["ctx"]
    template = ctx.jenv.get_template("trace.html")
    async with check_params(request, TaskIdParams) as params:
        if request.path.startswith("/trace-running"):
            trace_data = ctx.monitor.format_running_task_stack(params.task_id)
        elif request.path.startswith("/trace-terminated"):
            trace_data = ctx.monitor.format_terminated_task_stack(params.task_id)
        else:
            raise RuntimeError("should not reach here")
        output = template.render(
            navigation=nav_menus,
            page={
                "title": f"Task trace for {params.task_id}",
            },
            trace_data=trace_data,
        )
        return web.Response(body=output, content_type="text/html")


async def get_version(request: web.Request) -> web.Response:
    return web.json_response(
        data={
            "value": version("aiomonitor"),
        }
    )


async def get_task_count(request: web.Request) -> web.Response:
    async with check_params(request, TaskTypeParams) as params:
        ctx: WebUIContext = request.app["ctx"]
        if params.task_type == TaskTypes.RUNNING:
            count = len(asyncio.all_tasks(ctx.monitor._monitored_loop))
        elif params.task_type == TaskTypes.TERMINATED:
            count = len(ctx.monitor._terminated_history)
        else:
            raise RuntimeError("should not reach here")
        return web.json_response(
            data={
                "value": count,
            }
        )


async def get_live_task_list(request: web.Request) -> web.Response:
    ctx: WebUIContext = request.app["ctx"]
    async with check_params(request, ListFilterParams) as params:
        tasks = ctx.monitor.format_running_task_list(
            params.filter,
            params.persistent,
        )
        return web.json_response(
            data={
                "tasks": [
                    {
                        "task_id": t.task_id,
                        "state": t.state,
                        "name": t.name,
                        "coro": t.coro,
                        "created_location": t.created_location,
                        "since": t.since,
                        "is_root": t.created_location == "-",
                    }
                    for t in tasks
                ]
            }
        )


async def get_terminated_task_list(request: web.Request) -> web.Response:
    ctx: WebUIContext = request.app["ctx"]
    async with check_params(request, ListFilterParams) as params:
        tasks = ctx.monitor.format_terminated_task_list(
            params.filter,
            params.persistent,
        )
        return web.json_response(
            data={
                "tasks": [
                    {
                        "task_id": t.task_id,
                        "name": t.name,
                        "coro": t.coro,
                        "started_since": t.started_since,
                        "terminated_since": t.terminated_since,
                    }
                    for t in tasks
                ]
            }
        )


async def cancel_task(request: web.Request) -> web.Response:
    ctx: WebUIContext = request.app["ctx"]
    async with check_params(request, TaskIdParams) as params:
        try:
            coro_repr = await ctx.monitor.cancel_monitored_task(params.task_id)
            return web.json_response(
                data={
                    "msg": f"Successfully cancelled {params.task_id}",
                    "detail": coro_repr,
                },
            )
        except ValueError as e:
            return web.json_response(
                status=404,
                data={"msg": repr(e)},
            )


async def init_webui(monitor: Monitor) -> web.Application:
    jenv = Environment(
        loader=PackageLoader("aiomonitor.webui"), autoescape=select_autoescape()
    )
    app = web.Application()
    app["ctx"] = WebUIContext(
        monitor=monitor,
        jenv=jenv,
    )
    app.router.add_route("GET", "/", show_list_page)
    app.router.add_route("GET", "/about", show_about_page)
    app.router.add_route("GET", "/trace-running", show_trace_page)
    app.router.add_route("GET", "/trace-terminated", show_trace_page)
    app.router.add_route("GET", "/api/version", get_version)
    app.router.add_route("POST", "/api/task-count", get_task_count)
    app.router.add_route("POST", "/api/live-tasks", get_live_task_list)
    app.router.add_route("POST", "/api/terminated-tasks", get_terminated_task_list)
    app.router.add_route("DELETE", "/api/task", cancel_task)
    app.router.add_static("/static", Path(__file__).parent / "static")
    return app



import json
from abc import ABCMeta, abstractmethod
from contextlib import asynccontextmanager as actxmgr
from typing import AsyncIterator, Type, TypeVar

import trafaret as t
from aiohttp import web


class APIParams(metaclass=ABCMeta):
    @classmethod
    @abstractmethod
    def get_checker(cls) -> t.Trafaret:
        raise NotImplementedError

    @classmethod
    def check(cls, value):
        checker = cls.get_checker()
        data = checker.check(value)
        return cls(**data)  # assumes dataclass-like init with kwargs


T_APIParams = TypeVar("T_APIParams", bound=APIParams)


@actxmgr
async def check_params(
    request: web.Request,
    checker: Type[T_APIParams],
) -> AsyncIterator[T_APIParams]:
    try:
        if request.method in ("GET", "DELETE"):
            params = checker.check(request.query)
        else:
            body = await request.post()
            params = checker.check(body)
        yield params
    except t.DataError as e:
        error_data = e.as_dict()
        if isinstance(error_data, str):
            detail = error_data
        else:
            detail = "\n".join(v for v in error_data.values())
        raise web.HTTPBadRequest(
            content_type="application/json",
            body=json.dumps({"msg": "Invalid parameters", "detail": detail}),
        ) from None
    except Exception as e:
        raise web.HTTPInternalServerError(
            content_type="application/json",
            body=json.dumps({"msg": "Internal server error", "detail": repr(e)}),
        ) from e



# NOTE: This example requires Python 3.11 or higher.
# Use `requirements-dev.txt` to install dependencies.

from __future__ import annotations

import argparse
import asyncio
import contextlib
import enum
import logging
import weakref

import uvloop
from aiohttp import web

import aiomonitor
from aiomonitor.task import preserve_termination_log


class LoopImpl(enum.StrEnum):
    ASYNCIO = "asyncio"
    UVLOOP = "uvloop"


async def inner2() -> None:
    await asyncio.sleep(100)


async def inner1(tg: asyncio.TaskGroup) -> None:
    t = tg.create_task(inner2())
    await t


@preserve_termination_log
async def timer_loop() -> None:
    async def interval_work():
        await asyncio.sleep(3.0)

    scope: weakref.WeakSet[asyncio.Task] = weakref.WeakSet()
    try:
        while True:
            await asyncio.sleep(1.0)
            if len(scope) < 10:
                t = asyncio.create_task(interval_work(), name="interval-worker")
                scope.add(t)
    except asyncio.CancelledError:
        for t in {*scope}:
            with contextlib.suppress(asyncio.CancelledError):
                t.cancel()
                await t


async def simple(request: web.Request) -> web.Response:
    log = logging.getLogger("demo.aiohttp_server.simple")
    log.info("Start sleeping")
    async with asyncio.TaskGroup() as tg:
        tg.create_task(inner1(tg))
    log.info("Finished sleeping")
    return web.Response(text="Simple answer")


async def main(
    *,
    history_limit: int,
    num_timers: int,
) -> None:
    loop = asyncio.get_running_loop()
    log_timer_loop = logging.getLogger("demo.timer_loop")
    log_aiohttp_server = logging.getLogger("demo.aiohttp_server")
    with aiomonitor.start_monitor(
        loop,
        hook_task_factory=True,
        max_termination_history=history_limit,
    ):
        app = web.Application()
        app.router.add_get("/", simple)
        timer_loop_tasks = set()
        for idx in range(num_timers):
            timer_loop_tasks.add(
                asyncio.create_task(timer_loop(), name=f"timer-loop-{idx}")
            )
            log_timer_loop.info("started a timer loop (%d)", idx)
        app["timer_loop_tasks"] = timer_loop_tasks

        async def _shutdown_timer_loop_task(app: web.Application) -> None:
            for idx, t in enumerate(app["timer_loop_tasks"]):
                log_timer_loop.info("shutting down a timer loop (%d)", idx)
                t.cancel()
                await t

        app.on_cleanup.append(_shutdown_timer_loop_task)
        log_aiohttp_server.info(
            "started a sample aiohttp server: http://localhost:8090/"
        )
        await web._run_app(app, port=8090, host="localhost")


if __name__ == "__main__":
    logging.basicConfig()
    logging.getLogger("aiomonitor").setLevel(logging.DEBUG)
    logging.getLogger("demo").setLevel(logging.DEBUG)
    parser = argparse.ArgumentParser()
    parser.add_argument(
        "--loop",
        type=LoopImpl,
        choices=list(LoopImpl),
        default=LoopImpl.ASYNCIO,
        help="The event loop implementation to use [default: asyncio]",
    )
    parser.add_argument(
        "--history-limit",
        type=int,
        metavar="COUNT",
        default=10,
        help="Set the maximum number of task termination history [default: 10]",
    )
    parser.add_argument(
        "--num-timers",
        type=int,
        metavar="COUNT",
        default=5,
        help="Set the number of timer loop tasks to demonstrate persistent termination logs [default: 5]",
    )
    args = parser.parse_args()
    match args.loop:
        case LoopImpl.UVLOOP:
            uvloop.install()
        case _:
            pass
    try:
        asyncio.run(
            main(
                history_limit=args.history_limit,
                num_timers=args.num_timers,
            )
        )
    except KeyboardInterrupt:
        pass



import asyncio

import uvloop
from aiohttp import web

import aiomonitor


async def simple(request: web.Request) -> web.Response:
    await asyncio.sleep(10)
    await asyncio.sleep(10)
    return web.Response(text="Simple answer")


async def hello(request: web.Request) -> web.StreamResponse:
    resp = web.StreamResponse()
    name = request.match_info.get("name", "Anonymous")
    answer = ("Hello, " + name).encode("utf8")
    resp.content_length = len(answer)
    resp.content_type = "text/plain"
    await resp.prepare(request)
    await asyncio.sleep(100)
    await resp.write(answer)
    await resp.write_eof()
    return resp


async def main() -> None:
    loop = asyncio.get_running_loop()
    app = web.Application()
    app.router.add_get("/simple", simple)
    app.router.add_get("/hello/{name}", hello)
    app.router.add_get("/hello", hello)
    host, port = "localhost", 8090
    with aiomonitor.start_monitor(loop):
        web.run_app(app, port=port, host=host)


if __name__ == "__main__":
    uvloop.install()
    try:
        asyncio.run(main())
    except KeyboardInterrupt:
        pass



import asyncio

import click

import aiomonitor
from aiomonitor.termui.commands import (
    auto_async_command_done,
    auto_command_done,
    print_fail,
    print_ok,
)


@aiomonitor.monitor_cli.command(name="hello")
@auto_command_done
def do_hello(ctx: click.Context) -> None:
    """
    Command extension example
    """
    # To send outputs to the telnet client,
    # you should use either:

    click.echo("Hello, world!")

    # or:

    from aiomonitor.termui.commands import current_stdout

    stdout = current_stdout.get()
    print("Hello, world, again!", file=stdout)

    # or:

    from prompt_toolkit import print_formatted_text
    from prompt_toolkit.formatted_text import FormattedText

    print_formatted_text(
        FormattedText(
            [
                ("ansibrightblue", "Hello, "),
                ("ansibrightyellow", "world, "),
                ("ansibrightmagenta", "with color!"),
            ]
        )
    )

    # or:

    print_ok("Hello, world, success!")
    print_fail("Hello, world, failure!")


@aiomonitor.monitor_cli.command(name="hello-async")
def do_async_hello(ctx: click.Context) -> None:
    """
    Command extension example (async)
    """

    @auto_async_command_done
    async def _do_async_hello(ctx: click.Context):
        await asyncio.sleep(1)
        click.echo("Hello, world, after sleep!")

    asyncio.create_task(_do_async_hello(ctx))


async def main():
    loop = asyncio.get_running_loop()
    with aiomonitor.start_monitor(loop=loop) as monitor:
        print("Now you can connect with:")
        print(f"  python -m aiomonitor.cli -H {monitor.host} -p {monitor.port}")
        print("or")
        print(f"  telnet {monitor.host} {monitor.port}  # Linux only")
        # run forever until interrupted
        while True:
            await asyncio.sleep(60)


if __name__ == "__main__":
    try:
        asyncio.run(main())
    except KeyboardInterrupt:
        pass



import argparse
import asyncio

from .monitor import MONITOR_HOST, MONITOR_TERMUI_PORT
from .telnet import TelnetClient


async def async_monitor_client(host: str, port: int) -> None:
    async with TelnetClient(host, port) as client:
        await client.interact()


def monitor_client(host: str, port: int) -> None:
    try:
        asyncio.run(async_monitor_client(host, port))
    except KeyboardInterrupt:
        pass


def main() -> None:
    parser = argparse.ArgumentParser()
    parser.add_argument(
        "-H",
        "--host",
        dest="monitor_host",
        default=MONITOR_HOST,
        type=str,
        help="monitor host ip",
    )
    parser.add_argument(
        "-p",
        "--port",
        dest="monitor_port",
        default=MONITOR_TERMUI_PORT,
        type=int,
        help="monitor port number",
    )
    args = parser.parse_args()
    monitor_client(args.monitor_host, args.monitor_port)


if __name__ == "__main__":
    main()



import asyncio

import aiotools  # FIXME: replace with asyncio.TaskGroup in Python 3.11

import aiomonitor


async def inner2():
    await asyncio.sleep(100)


async def inner1():
    t = asyncio.create_task(inner2())
    await t


async def main():
    loop = asyncio.get_running_loop()
    with aiomonitor.start_monitor(loop, hook_task_factory=True):
        print("Starting...")
        async with aiotools.TaskGroup() as tg:
            for _ in range(10):
                tg.create_task(inner1())
        print("Done")


if __name__ == "__main__":
    try:
        asyncio.run(main())
    except KeyboardInterrupt:
        pass



from __future__ import annotations

import asyncio
import contextlib
import contextvars
import functools
import io
import sys
import unittest.mock
from typing import Sequence

import click
import pytest
from prompt_toolkit.application import create_app_session
from prompt_toolkit.input import create_pipe_input
from prompt_toolkit.output import DummyOutput

import aiomonitor.termui.commands
from aiomonitor import Monitor, start_monitor
from aiomonitor.termui.commands import (
    auto_command_done,
    command_done,
    current_monitor,
    current_stdout,
    custom_help_option,
    monitor_cli,
    print_ok,
)


@contextlib.contextmanager
def monitor_common():
    def make_baz():
        return "baz"

    locals_ = {"foo": "bar", "make_baz": make_baz}
    # In the tests, we reuse the pytest's event loop as the monitored loop.
    # Because of this, all cross-loop coroutine invocations should use the following
    # pattern in both tests and the monitor/termui implementation:
    # > fut = asyncio.wrap_future(asyncio.run_coroutine_threadsafe(...))
    # > await fut
    test_loop = asyncio.get_running_loop()
    mon = Monitor(test_loop, locals=locals_)
    with mon:
        yield mon


@pytest.fixture
async def monitor(request, event_loop):
    with monitor_common() as monitor_instance:
        yield monitor_instance


def get_task_ids(event_loop):
    return [id(t) for t in asyncio.all_tasks(loop=event_loop)]


class BufferedOutput(DummyOutput):
    def __init__(self) -> None:
        self._buffer = io.StringIO()

    def write(self, data: str) -> None:
        self._buffer.write(data)

    def write_raw(self, data: str) -> None:
        self._buffer.write(data)


async def invoke_command(
    monitor: Monitor,
    args: Sequence[str],
) -> str:
    dummy_stdout = BufferedOutput()
    current_monitor_token = current_monitor.set(monitor)
    current_stdout_token = current_stdout.set(dummy_stdout._buffer)

    async def _ui_create_event() -> asyncio.Event:
        return asyncio.Event()

    fut = asyncio.run_coroutine_threadsafe(_ui_create_event(), monitor._ui_loop)
    command_done_event: asyncio.Event = await asyncio.wrap_future(fut)
    command_done_token = command_done.set(command_done_event)
    try:
        with unittest.mock.patch.object(
            aiomonitor.termui.commands,
            "print_formatted_text",
            functools.partial(
                aiomonitor.termui.commands.print_formatted_text, output=dummy_stdout
            ),
        ):
            ctx = contextvars.copy_context()
            ctx.run(
                monitor_cli.main,
                args,
                prog_name="",
                obj=monitor,
                standalone_mode=False,  # type: ignore
            )
            # If Click raises UsageError before running the command,
            # there will be no one to set command_done_event.
            # In this case, the error is propagated to the upper stack
            # immediately here.
            fut = asyncio.run_coroutine_threadsafe(
                command_done_event.wait(),  # type: ignore
                monitor._ui_loop,
            )
            await asyncio.wrap_future(fut)
    finally:
        command_done.reset(command_done_token)
        current_stdout.reset(current_stdout_token)
        current_monitor.reset(current_monitor_token)
    with contextlib.closing(dummy_stdout._buffer):
        return dummy_stdout._buffer.getvalue()


@pytest.fixture(params=[True, False], ids=["console:True", "console:False"])
def console_enabled(request):
    return request.param


@pytest.mark.asyncio
async def test_ctor(event_loop, console_enabled):
    with Monitor(event_loop, console_enabled=console_enabled):
        await asyncio.sleep(0.01)
    with start_monitor(event_loop, console_enabled=console_enabled) as m:
        await asyncio.sleep(0.01)
    assert m.closed

    m = Monitor(event_loop, console_enabled=console_enabled)
    m.start()
    try:
        await asyncio.sleep(0.01)
    finally:
        m.close()
        m.close()  # make sure call is idempotent
    assert m.closed

    m = Monitor(event_loop, console_enabled=console_enabled)
    m.start()
    with m:
        await asyncio.sleep(0.01)
    assert m.closed

    # make sure that monitor inside async func can exit correctly
    with Monitor(event_loop, console_enabled=console_enabled):
        await asyncio.sleep(0.01)


@pytest.mark.asyncio
async def test_basic_monitor(event_loop, monitor: Monitor):
    resp = await invoke_command(monitor, ["help"])
    assert "Commands:" in resp

    with pytest.raises(click.UsageError):
        await invoke_command(monitor, ["xxx"])

    resp = await invoke_command(monitor, ["ps"])
    assert "Task" in resp

    with pytest.raises(click.UsageError):
        await invoke_command(monitor, ["ps", "123"])

    resp = await invoke_command(monitor, ["signal", "name"])
    assert "Unknown signal" in resp

    resp = await invoke_command(monitor, ["stacktrace"])
    assert "self.run_forever()" in resp

    resp = await invoke_command(monitor, ["w", "123"])
    assert "No task 123" in resp

    resp = await invoke_command(monitor, ["where", "123"])
    assert "No task 123" in resp

    with pytest.raises(click.UsageError):
        await invoke_command(monitor, ["c", "123"])

    resp = await invoke_command(monitor, ["cancel", "123"])
    assert "Invalid or non-existent task ID" in resp

    resp = await invoke_command(monitor, ["ca", "123"])
    assert "Invalid or non-existent task ID" in resp


myvar = contextvars.ContextVar("myvar", default=42)


@pytest.mark.asyncio
async def test_monitor_task_factory(event_loop):
    async def do():
        await asyncio.sleep(0)
        myself = asyncio.current_task()
        assert myself is not None
        assert myself.get_name() == "mytask"

    with Monitor(event_loop, console_enabled=False, hook_task_factory=True):
        t = asyncio.create_task(do(), name="mytask")
        await t


@pytest.mark.skipif(
    sys.version_info < (3, 11),
    reason="The context argument of asyncio.create_task() is added in Python 3.11",
)
@pytest.mark.asyncio
async def test_monitor_task_factory_with_context():
    ctx = contextvars.Context()
    # This context is bound at the outermost scope,
    # and inside it the initial value of myvar is kept intact.

    async def do():
        await asyncio.sleep(0)
        assert myvar.get() == 42  # we are referring the outer context
        myself = asyncio.current_task()
        assert myself is not None
        assert myself.get_name() == "mytask"

    myvar.set(99)  # override in the current task's context
    event_loop = asyncio.get_running_loop()
    with Monitor(event_loop, console_enabled=False, hook_task_factory=True):
        t = asyncio.create_task(do(), name="mytask", context=ctx)
        await t
    assert myvar.get() == 99


@pytest.mark.asyncio
async def test_cancel_where_tasks(
    monitor: Monitor,
) -> None:
    async def sleeper():
        await asyncio.sleep(100)  # xxx

    test_loop = monitor._monitored_loop
    t = test_loop.create_task(sleeper())
    t_id = id(t)
    await asyncio.sleep(0.1)

    task_ids = get_task_ids(test_loop)
    assert len(task_ids) > 0
    assert t_id in task_ids
    resp = await invoke_command(monitor, ["where", str(t_id)])
    assert "Task" in resp
    resp = await invoke_command(monitor, ["cancel", str(t_id)])
    assert "Cancelled task" in resp
    assert t.done()


@pytest.mark.asyncio
async def test_monitor_with_console(monitor: Monitor) -> None:
    with create_pipe_input() as pipe_input:
        stdout_buf = BufferedOutput()
        with create_app_session(input=pipe_input, output=stdout_buf):

            async def _interact():
                await asyncio.sleep(0.2)
                try:
                    pipe_input.send_text("await asyncio.sleep(0.1, result=333)\r\n")
                    pipe_input.flush()
                    await asyncio.sleep(0.1)
                    pipe_input.send_text("foo\r\n")
                    pipe_input.flush()
                    await asyncio.sleep(0.4)
                    resp = stdout_buf._buffer.getvalue()
                    assert "This console is running in an asyncio event loop." in resp
                    assert "333" in resp
                    assert "bar" in resp
                finally:
                    pipe_input.send_text("exit()\r\n")

            t = asyncio.create_task(_interact())
            await invoke_command(monitor, ["console"])
            await t
    # Check if we are back to the original shell.
    resp = await invoke_command(monitor, ["help"])
    assert "Commands" in resp


@pytest.mark.asyncio
async def test_custom_monitor_command(monitor: Monitor):
    @monitor_cli.command(name="something")
    @click.argument("arg")
    @custom_help_option
    @auto_command_done
    def do_something(ctx: click.Context, arg: str) -> None:
        print_ok(f"doing something with {arg}")

    resp = await invoke_command(monitor, ["something", "someargument"])
    assert "doing something with someargument" in resp
